structure Literal =
  struct
    datatype lit =
        LitInt of int
      | LitChar of char
      | LitString of string

    datatype binop = OpPlus | OpMinus | OpTimes | OpEquals | OpConcat
    datatype unop = OpPrint | OpStrToInt | OpIntToStr |
                    OpCharToInt | OpCharToStr | OpIntToChar

    datatype lit_ty =
        LitTyInt
      | LitTyChar
      | LitTyString

    fun lit_ty lit =
        case lit of
            LitInt _ => LitTyInt
          | LitChar _ => LitTyChar
          | LitString _ => LitTyString

    fun lit_ty_eq (lt1, lt2) =
        case (lt1, lt2) of
            (LitTyInt, LitTyInt) => true
          | (LitTyChar, LitTyChar) => true
          | (LitTyString, LitTyString) => true
          | (_, _) => false
     
  (* Returns input and ouptut types of a binary operator a
      representation of boolean and unit types in some language
      as well as a function that takes a literal to a type
      in that language, typically the TyLit constructor. *)
    fun binop_tys oper f tyBool tyUnit =
      case oper of
        OpPlus => (f LitTyInt, f LitTyInt, f LitTyInt)
      | OpMinus => (f LitTyInt, f LitTyInt, f LitTyInt)
      | OpTimes => (f LitTyInt, f LitTyInt, f LitTyInt)
      | OpEquals => (f LitTyInt, f LitTyInt, tyBool)
      | OpConcat => (f LitTyString, f LitTyString, f LitTyString)

  (* Returns input and ouptut types of a unary operator a
      representation of boolean and unit types in some language
      as well as a function that takes a literal to a type
      in that language, typically the TyLit constructor. *)
    fun unop_tys oper f tyBool tyUnit =
      case oper of
        OpPrint => (f LitTyString, tyUnit)
      | OpStrToInt => (f LitTyString, f LitTyInt)
      | OpIntToStr => (f LitTyInt, f LitTyString)
      | OpCharToInt => (f LitTyChar, f LitTyInt)
      | OpCharToStr => (f LitTyChar, f LitTyString)
      | OpIntToChar => (f LitTyInt, f LitTyChar)

    fun binop_to_string oper = 
      case oper of
        OpEquals => "equals"
      | OpPlus => "add"
      | OpMinus => "sub"
      | OpTimes => "mul"
      | OpConcat => "concat"
    
    fun unop_to_string oper =
      case oper of
        OpPrint => "print"
      | OpStrToInt => "str_to_int"
      | OpIntToStr => "int_to_str"
      | OpCharToInt => "chr_to_int"
      | OpCharToStr => "chr_to_str"
      | OpIntToChar => "int_to_chr"

  end
