structure ClosureConvert :> CLOSURE_CONVERT =
  struct
    structure Var = Variable

    structure IL4S = IL4Static
    structure Ctx = IL4S.Ctx

    type context = Ctx.context

    exception Convert of string
    exception Error of string

(* Conversion functions *)
    fun convert_kind k =
      case k of
        IL4.KndType => IL5.KndType
      | IL4.KndArrow(k1, k2) => IL5.KndArrow(convert_kind k1, convert_kind k2)
      | IL4.KndUnit => IL5.KndUnit
      | IL4.KndPair(k1, k2) => IL5.KndPair(convert_kind k1, convert_kind k2)

    fun convert_con con = raise Error "Unimplemented"

    fun convert_value G v =
      case v of
          IL4.ValVar x => IL5.ValVar x
        | IL4.ValLit l => IL5.ValLit l
        | IL4.ValProd vs => IL5.ValProd (map (convert_value G) vs)
        | IL4.ValSumInj(c, i, v) =>
            IL5.ValSumInj(convert_con c, i, convert_value G v)
        | IL4.ValTagInj(v1, v2) =>
            IL5.ValTagInj(convert_value G v1, convert_value G v2)
        | IL4.ValConApp(v, c) => IL5.ValConApp(convert_value G v, convert_con c)
        | IL4.ValRoll(c, v) => IL5.ValRoll(convert_con c, convert_value G v)
      (*
        | ValConAbs
        | ValFix
      *)

    and convert_exp G exp =
      case exp of
          IL4.ExHalt v => IL5.ExHalt(convert_value G v)
        | IL4.ExIfTagOf(v1, v2, (x, e1) , e2) =>
          let
            val c = case IL4S.whnf (IL4S.synth_value_con G v2) of
                        IL4.TyTag c => c
                      | _ => raise Convert "non-tag passed to iftagof"
            val (x', e1') = IL4.alpha_vary (x, e1) IL4.replace_exp_exp 
            val G' = Ctx.extend G (Ctx.DecExpCon(x', c))
            val e1'' = convert_exp G' e1'
            val v1' = convert_value G v1
            val v2' = convert_value G v2
            val e2' = convert_exp G e2
          in
            IL5.ExIfTagOf(v1', v2', (x', e1''), e2')
          end
        | IL4.ExLet(b, e) =>
          let
            val (x, x', c, b') = convert_bind G b
            val e' = IL4.replace_exp_exp x x' e
            val G' = Ctx.extend G (Ctx.DecExpCon(x', c))
            val e'' = convert_exp G' e'
          in
            IL5.ExLet(b', e'')
          end
      (*
        | ExApp
        | ExCase
      *)
    and convert_bind G bnd =
      case bnd of
           IL4.BndVal(x, v) =>
             let
               val x' = Var.freshen x
               val v' = convert_value G v
             in
              (x, x', IL4S.synth_value_con G v, IL5.BndVal(x', v'))
             end
         | IL4.BndUnOp(x, oper, v) =>
             let
               val x' = Var.freshen x
               val v' = convert_value G v
               val (_, c) =
                    Literal.unop_tys oper IL4.TyLit IL4.tyBool IL4.tyUnit
             in
              (x, x', c, IL5.BndUnOp(x', oper, v'))
             end
         | IL4.BndProj(x, i, v) =>
            let
              val x' = Var.freshen x
              val v' = convert_value G v
              val c = case IL4S.synth_value_con G v of
                           IL4.TyProd cs => cs
                         | _ => raise Convert "non product in projection"
            in
              (x, x', List.nth(c, i - 1), IL5.BndProj(x', i, v'))
            end
         | IL4.BndRef(x, v) =>
            let
              val x' = Var.freshen x
              val v' = convert_value G v
              val c = IL4S.synth_value_con G v
            in
              (x, x', IL4.TyRef c, IL5.BndRef(x', v'))
            end
         | IL4.BndGet(x, v) =>
            let
              val x' = Var.freshen x
              val v' = convert_value G v
              val c = case IL4S.whnf (IL4S.synth_value_con G v) of
                           IL4.TyRef c => c
                         | _ => raise Convert "non ref value in deref"
            in
              (x, x', c, IL5.BndGet(x', v'))
            end
        | IL4.BndNewTag(x, c) =>
            let
              val x' = Var.freshen x
            in
              (x, x', IL4.TyTag c, IL5.BndNewTag(x', convert_con c))
            end
        | IL4.BndUnroll(x, v) =>
            let
              val x' = Var.freshen x
              val v' = convert_value G v
              val c = (IL4S.whnf (IL4S.synth_value_con G v))
            in
              (x, x', IL4S.expand G c, IL5.BndUnroll(x', v'))
            end
      (*
        | BndBinOp
        | BndSet
       *)
  end
