(* Generates a converter top level given an:
      in language
      out language
      set of conversion functions
*)


functor ConvertTopFn(L : sig
                        structure In :
                          sig
                            type kind
                            type con
                            type term

                            exception Check of string
                            exception Error of string

                            val kind_to_layout : kind -> Layout.layout
                            val con_to_layout : con -> Layout.layout
                            val term_to_layout : term -> Layout.layout
                          end
                        structure Out :
                          sig
                            type kind
                            type con
                            type term

                            val kind_to_layout : kind -> Layout.layout
                            val con_to_layout : con -> Layout.layout
                            val term_to_layout : term -> Layout.layout
                          end
                        exception Convert of string
                        exception Error of string

                        val convert_kind : In.kind -> Out.kind
                        val convert_con : In.con -> Out.con
                        val convert_term : In.term -> Out.term
                       end) :> CONVERT_TOP
                            where type kind = L.In.kind
                              and type con = L.In.con
                              and type term = L.In.term =
  struct

    type kind = L.In.kind
    type con = L.In.con
    type term = L.In.term

    fun try f x =
        f x
        handle
            L.Convert string =>
                print ("conversion error: " ^ string ^ "\n")
          | L.Error string =>
                print ("conversion internal error: " ^ string ^ "\n")
          | L.In.Check string =>
                print ("Input Language Static type error: " ^ string ^ "\n")
          | L.In.Error string =>
                print ("Input Language internal error: " ^ string ^ "\n")
    
    fun convert_knd' knd =
        let val outknd = L.convert_kind knd
            val layout =
                Layout.align
                    [ Layout.seq [Layout.str "  |- ",
                                  L.In.kind_to_layout knd],
                      Layout.seq [Layout.str " ~~> ",
                                  L.Out.kind_to_layout outknd] ]
            (* weird... it seems to go over if you allow the default of 80 *)
            val s = Layout.tostringex 77 layout
        in
            print (s ^ "\n")
        end

    fun convert_con' con =
        let val outcon = L.convert_con con
            val layout =
                Layout.align
                    [ Layout.seq [Layout.str "  |- ",
                                  L.In.con_to_layout con],
                      Layout.seq [Layout.str " ~~> ",
                                  L.Out.con_to_layout outcon] ]
            (* weird... it seems to go over if you allow the default of 80 *)
            val s = Layout.tostringex 77 layout
        in
            print (s ^ "\n")
        end

    val $ = Layout.str
    fun var x = $(Variable.toString x)

    fun convert_term' term =
        let val term' =
                L.convert_term term
            val layout =
                Layout.align
                    [ Layout.seq [Layout.str "  |- ",
                                  L.In.term_to_layout term],
                      Layout.seq [Layout.str " ~~> ",
                                  L.Out.term_to_layout term'] ]
            val s = Layout.tostringex 77 layout
        in
            print (s ^ "\n")
        end

    fun convert_kind knd = try convert_knd' knd
    fun convert_con con = try convert_con' con
    fun convert_term term = try convert_term' term

  end
