structure CPSConvert =
  struct
    
    structure IL3S = IL3Static
    structure Ctx = IL3S.Ctx
    structure Var = Variable
    
    type context = Ctx.context

    exception Convert of string
    exception Error of string


    fun convert_kind knd = case knd of
          IL3.KndType => IL4.KndType
        | IL3.KndArrow (k1, k2) =>
            IL4.KndArrow(convert_kind k1, convert_kind k2)
        | IL3.KndUnit => IL4.KndUnit
        | IL3.KndPair (k1, k2) =>
            IL4.KndPair(convert_kind k1, convert_kind k2)

    fun convert_con con = raise Error "Unimplemented"


    and convert_exp G exp = case exp of
          IL3.ExVar x =>
            let
              val con = case Ctx.lookup_expvar G x of
                            NONE => raise Convert
                              ("Unbound var " ^ (Var.toString x))
                          | SOME con => con
              val k = Var.newNamed "k"
              val k_exn = Var.newNamed "k_exn"
              val rexp = IL4.ExApp(IL4.ValVar k, IL4.ValVar x)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExLit l =>
            let
              val con = IL3.TyLit(Literal.lit_ty l)
              val k = Var.newNamed "k"
              val k_exn = Var.newNamed "k_exn"
              val rexp = IL4.ExApp(IL4.ValVar k, IL4.ValLit l)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExUnOp(oper, exp) =>
            let
              val (c, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val (_, con) =
                    Literal.unop_tys oper IL3.TyLit IL3.tyBool IL3.tyUnit
              val f =
                IL4.valLam (x, convert_con c,
                  IL4.ExLet(IL4.BndUnOp(y, oper, xvar),
                    IL4.ExApp(kvar, yvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExRef exp =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val c = convert_con c1
              val f = IL4.valLam (x, c, IL4.ExLet(IL4.BndRef(y, xvar),
                  IL4.ExApp(kvar, yvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (IL3.TyRef c1, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExGet exp =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val con =
                  case IL3S.whnf c1 of
                    IL3.TyRef con => con
                  | _ => raise Convert "get from non ref type"
              val c = convert_con c1
              val f = IL4.valLam (x, c, IL4.ExLet(IL4.BndGet(y, xvar),
                  IL4.ExApp(kvar, yvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExSet(exp1, exp2) =>
            let
              val k = Var.newNamed "k"
              val k_exn = Var.newNamed "k_exn"
              val (c1, {k = k1, k_exn = k_exn1}, e1) = convert_exp G exp1
              val (c2, {k = k2, k_exn = k_exn2}, e2) = convert_exp G exp2
              val x1 = Var.newNamed "x1"
              val x2 = Var.newNamed "x2"
              val y = Var.newNamed "y"
              val x1val = IL4.ValVar x1
              val x2val = IL4.ValVar x2
              val yval = IL4.ValVar y
              val kval = IL4.ValVar k
              val k_exnval = IL4.ValVar k_exn
              val f2bod =
                IL4.ExLet(IL4.BndSet(y, x1val, x2val), IL4.ExApp(kval, yval))
              val f2 = IL4.valLam(x2, convert_con c2, f2bod)
              val f1bod = IL4.ExLet(IL4.BndVal(k2, f2),
                    IL4.ExLet(IL4.BndVal(k_exn2, k_exnval), e2))
              val f1 = IL4.valLam(x1, convert_con c1, f1bod)
              val rexp = IL4.ExLet(IL4.BndVal(k1, f1),
                    IL4.ExLet(IL4.BndVal(k_exn1, k_exnval), e1))
            in
              (IL3.tyUnit, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExRoll (con, exp) =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val c = convert_con con
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val f = IL4.valLam(x, convert_con c1,
                IL4.ExApp(kvar, IL4.ValRoll(c, xvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExUnroll exp =>
            let
              val (c, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val con = IL3S.expand G c
              val f =
                IL4.valLam (x, convert_con c, IL4.ExLet(IL4.BndUnroll(y, xvar),
                  IL4.ExApp(kvar, yvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExProd exps =>
            let
              val vars = map (fn _ => Var.newNamed "a") exps
              val varvals = map IL4.ValVar vars
              val k = Var.newNamed "k"
              val k_exn = Var.newNamed "k_exn"
              val kvar = IL4.ValVar k
              val k_exnvar = IL4.ValVar k_exn
              val outs = map (convert_exp G) exps
              val cons = map (fn (c, _, _) => c) outs
              fun wrap_part ((x, (con, {k = ki, k_exn = k_exni}, expi)), acc) =
                  IL4.ExLet(IL4.BndVal(ki, IL4.valLam(x, convert_con con, acc)),
                    IL4.ExLet(IL4.BndVal(k_exni, k_exnvar), expi))
              val nouts = ListPair.zip(vars, outs)
              val rexp =
                foldr wrap_part (IL4.ExApp(kvar, IL4.ValProd varvals)) nouts
            in
              (IL3.TyProd cons, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExProj (i, exp) =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val kvar = IL4.ValVar k
              val k1var = IL4.ValVar k1
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val cons =
                  case IL3S.whnf c1 of
                    IL3.TyProd cons => cons
                  | _ => raise Convert "projection from non product type"
              val c = convert_con c1
              val f = IL4.valLam (x, c, IL4.ExLet(IL4.BndProj(y, i, xvar),
                  IL4.ExApp(kvar, yvar)))
              val rexp = IL4.ExLet(IL4.BndVal(k1, f), e1)
            in
              (List.nth(cons, i - 1), {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExSumInj(csum, i, exp) =>
            let
              val cons =
                  case IL3S.whnf csum of
                    IL3.TySum cons => cons
                  | _ => raise Convert "injection with non sum type"
              val con = List.nth(cons, i - 1)
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val kvar = IL4.ValVar k
              val xvar = IL4.ValVar x
              val rexp = IL4.ExLet(IL4.BndVal(k1, IL4.valLam(x,
                convert_con c1, IL4.ExApp(kvar,
                  IL4.ValSumInj(convert_con csum, i, xvar)))), e1)
            in
              (csum, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExTagInj(exp1, exp2) =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp1
              val (c2, {k = k2, k_exn = k_exn2}, e2) = convert_exp G exp2
              val k = Var.newNamed "k"
              val x = Var.newNamed "x"
              val y = Var.newNamed "y"
              val xvar = IL4.ValVar x
              val yvar = IL4.ValVar y
              val kvar = IL4.ValVar k
              val k_exnvar = IL4.ValVar k_exn
              val f2bod = IL4.ExApp(kvar, IL4.ValTagInj(xvar, yvar))
              val f2 = IL4.valLam(y, convert_con c2, f2bod)
              val f1bod = IL4.ExLet(IL4.BndVal(k2, f2),
                    IL4.ExLet(IL4.BndVal(k_exn2, k_exnvar), e2))
              val f1 = IL4.valLam(x, convert_con c1, f1bod)
              val rexp = IL4.ExLet(IL4.BndVal(k1, f1), e1)
            in
              (IL3.TyTagged, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExNewTag con =>
            let
              val x = Var.newNamed "x"
              val k = Var.newNamed "k"
              val k_exn = Var.newNamed "k_exn"
              val xvar = IL4.ValVar x
              val kvar = IL4.ValVar k
              val c = convert_con con
              val rexp = IL4.ExLet(IL4.BndNewTag(x, c), IL4.ExApp(kvar, xvar))
            in
              (IL3.TyTag con, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExIfTagOf(exp1, exp2, exp3, exp4) =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp1
              val (c2, {k = k2, k_exn = k_exn2}, e2) = convert_exp G exp2
              val (c3, {k = k3, k_exn = k_exn3}, e3) = convert_exp G exp3
              val (c4, {k = k4, k_exn = k_exn4}, e4) = convert_exp G exp4
              val k = Var.newNamed "k"
              val v = Var.newNamed "v"
              val t = Var.newNamed "t"
              val x = Var.newNamed "x"
              val f = Var.newNamed "f"
              val kvar = IL4.ValVar k
              val vvar = IL4.ValVar v
              val tvar = IL4.ValVar t
              val xvar = IL4.ValVar x
              val fvar = IL4.ValVar f
              val k_exnvar = IL4.ValVar k_exn
              val tcase =
                IL4.ExLet(
                  IL4.BndVal(k3, IL4.valLam(f, convert_con c3, IL4.ExApp(fvar,
                      IL4.ValProd [kvar, k_exnvar, xvar]))),
                  IL4.ExLet(IL4.BndVal(k_exn3, k_exnvar), e3))
              val fcase =
                IL4.ExLet(IL4.BndVal(k4, kvar),
                      IL4.ExLet(IL4.BndVal(k_exn4, k_exnvar), e4))
              val itage = IL4.ExIfTagOf(vvar, tvar, (x, tcase), fcase)
              val rexp =
                IL4.ExLet(IL4.BndVal(k1,
                  IL4.valLam(v, convert_con c1,
                    IL4.ExLet(IL4.BndVal(k2,
                      IL4.valLam(t, convert_con c2, itage)),
                        IL4.ExLet(IL4.BndVal(k_exn2, k_exnvar), e2)))), e1)
            in
              (c4, {k = k, k_exn = k_exn}, rexp)
            end
        | IL3.ExLet(x, exp1, exp2) =>
            let
              val (c1, {k = k1, k_exn = k_exn}, e1) = convert_exp G exp1
              val G' = Ctx.extend G (Ctx.DecExpCon(x, c1))
              val (con, {k = k, k_exn = k_exn2}, e2) = convert_exp G' exp2
              val k_exnvar = IL4.ValVar k_exn
              val bind = IL4.valLam(x, convert_con c1,
                IL4.ExLet(IL4.BndVal(k_exn2, k_exnvar), e2))
              val rexp = IL4.ExLet(IL4.BndVal(k1, bind), e1)
            in
              (con,  {k = k, k_exn = k_exn}, rexp)
            end
     (*
       | ExBinOp
       | ExFix
       | ExApp
       | ExCase
       | ExHandle
       | ExRaise
       | ExConAbs
       | ExConApp 
     *)
  end
