structure IL3Examples =
  struct
    local
      open IL3 Literal
    in
      (* Constructors *)
      val arrow = TyArrow (tyInt, tyChar)
      val nested_arrow = TyArrow (TyArrow (tyInt, tyChar), tyString)
      val higher_arrow = TyArrow (tyInt, TyArrow (tyChar, tyString))

      local
          val cx = Variable.newNamed "a"
      in
      val forall =
          TyForall (cx, KndType,
                    TyArrow (ConVar cx, TyProd [ConVar cx, ConVar cx]))
      end

      local
          val cx = Variable.newNamed "a"
      in
      val bigcon =
          ConApp (ConAbs (cx, KndPair (KndType, KndUnit),
                          ConPair (ConProj1 (ConVar cx), ConProj2 (ConVar cx))),
                  ConPair (tyInt, ConUnit))
      end

      (* Values *)
      val const = exString "look Mom, no flow control!"
      local
          val x = Variable.newNamed "x"
      in
      val function = exLam (x, tyInt, tyInt, ExVar x)
      end

      (* Operations *)
      val add = ExBinOp (exInt 5, OpPlus, exInt 7)
      val eq = ExBinOp (exInt 10, OpEquals, exInt 12)
      val proj = ExProj (2, ExProd [exChar #"d", exInt 6])
                 (* take 2d6 due to sneak attack from low-level rogue *)

      (* Expressions *)
      local
          val f = Variable.newNamed "id"
          val x = Variable.newNamed "x"
      in
      val app = ExLet (f, exLam (x, tyString, tyString, ExVar x),
                       ExApp (ExVar f, exString "hello!"))
      end

      (* this one is big. *)
      local
          val x = Variable.newNamed "x"
      in
      val knot = exLam (x, tyBool, tyBool,
                       exIf tyBool (ExVar x, exBool false, exBool true))
      end

      (* this one is BIG! *)
      local
          val f = Variable.newNamed "f"
          val x = Variable.newNamed "x"
      in
      val fact =
          ExLet (f, ExFix (1, [(f, x, tyInt, tyInt,
                                exIf tyInt (ExBinOp (ExVar x,
                                                  OpEquals,
                                                  exInt 0),
                                            exInt 1,
                                            ExBinOp (ExVar x,
                                                  OpTimes,
                                                  ExApp (ExVar f,
                                                         ExBinOp (ExVar x,
                                                               OpMinus,
                                                               exInt 1)))))]),
                ExApp (ExVar f, ExApp (ExVar f, exInt 5)))
      end

      local
          val y = Variable.newNamed "y"
      in
      val letexp = ExLet (y, exInt 1, ExVar y)
      end

      local
          val tag = Variable.newNamed "tag"
          val exn = Variable.newNamed "exn"
          val y = Variable.newNamed "y"
      in
      val letexp2 =
          ExLet (tag, ExNewTag tyUnit,
                 ExLet (y, ExRaise (tyInt, ExTagInj (ExVar tag, exUnit)), ExVar y))
      end

      local
          val exn = Variable.newNamed "exn"
          val tag = Variable.newNamed "tag"
      in
      val handleexp =
          ExLet (tag, ExNewTag tyUnit,
                 ExHandle (ExTagInj (ExVar tag, exUnit), (exn, ExVar exn)))
      end

      local
          val exn = Variable.newNamed "exn"
          val tag = Variable.newNamed "tag"
      in
      val handleexp2 =
          ExLet (tag, ExNewTag tyUnit,
                 ExHandle (ExRaise (TyTagged, ExTagInj (ExVar tag, exUnit)),
                           (exn, ExVar exn)))
      end

      local
          val xref = Variable.newNamed "xref"
          val u = Variable.newNamed "_"
      in
      val counter =
          ExLet (xref, ExRef (exInt 0),
                 ExProd [exLam (u, tyUnit, tyUnit,
                                ExSet (ExVar xref, ExBinOp (ExGet (ExVar xref),
                                                         OpPlus, exInt 1))),
                         exLam (u, tyUnit, tyInt, ExGet (ExVar xref))])
      end

      (* hooray for copying test cases from previous assignments! *)
      local
          val cx  = Variable.newNamed "a"
          val c_mu = ConMu (cx, kndProd [KndType],
                                conProd [TyArrow (conProj (1, ConVar cx),
                                                  conProj (1, ConVar cx))])
          val x = Variable.newNamed "x"
          val cx2 = Variable.newNamed "b"
      in
      val rollerball =
          ExUnroll (ExRoll (conProj (1, c_mu),
                            ExConApp (ExConAbs (cx2, KndType,
                                                exLam (x, ConVar cx2,
                                                       ConVar cx2, ExVar x)),
                                      conProj (1, c_mu))))
      end
      val helloworld = ExUnOp(Literal.OpPrint, exString "hello world\n")
      val con_tests = [ arrow, nested_arrow, higher_arrow, forall, bigcon ]
      
      val exp_tests =
        [ const, function, add, eq, proj, app, knot, fact, letexp, letexp2,
          handleexp, handleexp2, counter, rollerball ]
    end
end
