structure IL5 :> IL5 =
  struct
    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    (* XXX could use datatype copying instead...
       datatype kind = datatype IL3.kind *)
    datatype kind = datatype IL4.kind

    datatype con =
        ConVar of convar
      | ConAbs of convar * kind * con
      | ConApp of con * con
      | ConUnit
      | ConPair of con * con
      | ConProj1 of con
      | ConProj2 of con
      | ConMu of convar * kind * con (* convar bound in con, kind = KndType^n *)

      | TyLit of Literal.lit_ty
      | TyCont of con (* instead of TyArrow *)
      | TyRef of con
      | TyTagged
      | TyTag of con
      | TySum of con list
      | TyProd of con list
      | TyForall of convar * kind * con (* convar bound in con *)
      | TyExists of convar * kind * con

    datatype value =
        ValVar of expvar
      | ValLit of Literal.lit
      | ValFix of int * fbnd list (* includes projection *)
      | ValRoll of con * value
      | ValProd of value list
      | ValSumInj of con * int * value
      | ValTagInj of value * value
      | ValConAbs of convar * kind * value (* convar bound in value *)
      | ValConApp of value * con
      | ValPack of (con * value) * con

    and exp =
        ExHalt of value
      | ExApp of value * value
      | ExCase of value * (expvar * exp) list
      | ExIfTagOf of value * value * (expvar * exp) * exp
      | ExLet of bnd * exp (* expvar bound in exp2 *)
      | ExUnpack of convar * expvar * value * exp

    and bnd =
        BndVal of expvar * value
      | BndUnOp of expvar * Literal.unop * value
      | BndBinOp of expvar * value * Literal.binop * value
      | BndRef of expvar * value
      | BndGet of expvar * value
      | BndSet of expvar * value * value
      | BndUnroll of expvar * value
      | BndProj of expvar * int * value
      | BndNewTag of expvar * con

    withtype fbnd = (expvar * expvar * con * exp)
              (* fix f       (x      : t) => e   end *)

    (*** mapping ***)

    (* map_con : (convar -> con)
              -> (convar * con -> convar * con)
              -> con -> con *)
    fun map_con fconvar fconvarcon con =
        let fun fcon con = map_con fconvar fconvarcon con
            fun fconvarkindcon (x, k, c) =
                    let val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
        in
            case con of
                ConVar x => fconvar x
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConUnit => ConUnit
              | ConPair (c1, c2) => ConPair (fcon c1, fcon c2)
              | ConProj1 c => ConProj1 (fcon c)
              | ConProj2 c => ConProj2 (fcon c)
              | ConMu xkc => ConMu (fconvarkindcon xkc)
              | TyLit lt => con
              | TyCont c => TyCont (fcon c)
              | TyRef c => TyRef (fcon c)
              | TyTagged => con
              | TyTag c => TyTag (fcon c)
              | TySum cons => TySum (map fcon cons)
              | TyProd cons => TyProd (map fcon cons)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
              | TyExists xkc => TyExists (fconvarkindcon xkc)
        end

    (* map_value : (expvar -> value)
                -> (fbnd list -> fbnd list)
                -> (convar * value -> convar * value)
                -> (con -> con)
                -> value -> value *)
    fun map_value fexpvar
                  ffbnds
                  fconvarvalue
                  fcon
                  value =
        let val fvalue = map_value fexpvar ffbnds fconvarvalue fcon
            fun fconvarkindvalue (x, k, e) =
                let val (x, e) = fconvarvalue (x, e)
                in
                    (x, k, e)
                end
        in
            case value of
                ValLit _ => value
              | ValVar x => fexpvar x
              | ValFix (i, fbnds) => ValFix (i, ffbnds fbnds)
              | ValProd vals => ValProd (map fvalue vals)
              | ValRoll (c, v) => ValRoll (fcon c, fvalue v)
              | ValSumInj (c, i, v) => ValSumInj (fcon c, i, fvalue v)
              | ValTagInj (v1, v2) => ValTagInj (fvalue v1, fvalue v2)
              | ValConAbs xkv => ValConAbs (fconvarkindvalue xkv)
              | ValConApp (v, c) => ValConApp (fvalue v, fcon c)
              | ValPack((c1, v), c2) => ValPack((fcon c1, fvalue v), fcon c2)
        end

    (* map_exp : (expvar * exp -> expvar * exp)
              -> (convar * expvar * exp -> convar * expvar * exp
              -> (bnd * exp -> bnd * exp)
              -> (value -> value)
              -> exp -> exp *)
    fun map_exp fexpvarexp
                fconexpvarexp
                fbndexp
                fvalue
                exp =
        let val fexp = map_exp fexpvarexp fconexpvarexp fbndexp fvalue
        in
            case exp of
                ExHalt v => ExHalt (fvalue v)
              | ExApp (v1, v2) => ExApp (fvalue v1, fvalue v2)
              | ExCase (v, xes) => ExCase (fvalue v, map fexpvarexp xes)
              | ExIfTagOf (v1, v2, xe1, e2) =>
                    ExIfTagOf (fvalue v1, fvalue v2, fexpvarexp xe1, fexp e2)
              | ExLet bndexp => ExLet (fbndexp bndexp)
              | ExUnpack (a, x, v, e) =>
                  let
                    val (a', x', e') = fconexpvarexp (a, x, e)
                    val v' = fvalue v
                  in
                    ExUnpack(a', x', v', e')
                  end
        end

    (* map_bnd : (value -> value)
              -> (con -> con)
              -> bnd -> bnd *)
    fun map_bnd fvalue fcon bnd =
        let val fbnd = map_bnd fvalue fcon
        in
            case bnd of
                BndVal (x, v) => BndVal (x, fvalue v)
              | BndUnOp(x, oper, v) => BndUnOp(x, oper, fvalue v)
              | BndBinOp (x, v1, oper, v2) =>
                  BndBinOp (x, fvalue v1, oper, fvalue v2)
              | BndRef (x, v) => BndRef (x, fvalue v)
              | BndGet (x, v) => BndGet (x, fvalue v)
              | BndSet (x, v1, v2) => BndSet (x, fvalue v1, fvalue v2)
              | BndUnroll (x, v) => BndUnroll (x, fvalue v)
              | BndProj (x, i, v) => BndProj (x, i, fvalue v)
              | BndNewTag (x, c) => BndNewTag (x, fcon c)
        end

    fun bnd_var bnd =
        case bnd of
            BndVal (x, _) => x
          | BndUnOp (x, _, _) => x
          | BndBinOp (x, _, _, _) => x
          | BndRef (x, _) => x
          | BndGet (x, _) => x
          | BndSet (x, _, _) => x
          | BndUnroll (x, _) => x
          | BndProj (x, _, _) => x
          | BndNewTag (x, _) => x


    (*** replacement of convars ***)

    fun replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                con

    fun replace_con_value convar convar' value =
        map_value (fn x => ValVar x)
                  (fn fbnds =>
                    map (fn (f, x, c, e) =>
                            (f, x, replace_con_con convar convar' c,
                                   replace_con_exp convar convar' e))
                        fbnds)
                  (fn (cx, v) =>
                    if Variable.eq (cx, convar) then
                        (cx, v)
                    else
                        (cx, replace_con_value convar convar' v))
                  (replace_con_con convar convar')
                  value

    and replace_con_exp convar convar' exp =
        map_exp (fn (ex, e) => (ex, replace_con_exp convar convar' e))
                (fn (cx, ex, e) =>
                  if Variable.eq (cx, convar) then
                    (cx, ex, e)
                  else
                    (cx, ex, replace_con_exp convar convar' e))
                (fn (b, e) => (replace_con_bnd convar convar' b,
                               replace_con_exp convar convar' e))
                (replace_con_value convar convar')
                exp

    and replace_con_bnd convar convar' bnd =
        map_bnd (replace_con_value convar convar')
                (replace_con_con convar convar')
                bnd

    fun replace_exp_value expvar expvar' value =
        map_value (fn x =>
                    if Variable.eq (x, expvar) then
                        ValVar expvar'
                    else
                        ValVar x)
                  (fn fbnds =>
                    if List.exists
                        (fn (f, _, _, _) => Variable.eq (f, expvar))
                        fbnds
                    then
                        (* expvar is bound in entire phrase, do nothing *)
                        fbnds
                    else
                        map (fn (f, x, c1, e) =>
                                let val e =
                                    if Variable.eq (x, expvar) then
                                        (* expvar is bound in e, do nothing *)
                                        e
                                    else
                                        replace_exp_exp expvar expvar' e
                                in
                                    (f, x, c1, e)
                                end)
                            fbnds)
                  (fn (cx, v) => (cx, replace_exp_value expvar expvar' v))
                  (fn c => c) (* expvars do not appear in cons *)
                  value

    and replace_exp_bnd expvar expvar' bnd =
        map_bnd (replace_exp_value expvar expvar')
                (fn c => c) (* expvars do not appear in cons *)
                bnd

    and replace_exp_exp expvar expvar' exp =
        map_exp (fn (ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (ex, e)
                    else
                        (ex, replace_exp_exp expvar expvar' e))
                (fn (cx, ex, e) =>
                    if Variable.eq (ex, expvar) then
                        (cx, ex, e)
                    else
                        (cx, ex, replace_exp_exp expvar expvar' e))
                (fn (b, e) =>
                    let val b = replace_exp_bnd expvar expvar' b
                        val x = bnd_var b
                    in
                        if Variable.eq (x, expvar) then
                            (b, e)
                        else
                            (b, replace_exp_exp expvar expvar' e)
                    end)
                (replace_exp_value expvar expvar')
                exp


    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end

    fun alpha_vary_bnd_exp (bnd, exp) =
        case bnd of
            BndVal (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndVal (x, v), exp)
                end
          | BndUnOp (x, oper, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndUnOp (x, oper, v), exp)
                end
          | BndBinOp (x, v1, oper, v2) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndBinOp (x, v1, oper, v2), exp)
                end
          | BndRef (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndRef (x, v), exp)
                end
          | BndGet (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndGet (x, v), exp)
                end
          | BndSet (x, v1, v2) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndSet (x, v1, v2), exp)
                end
          | BndUnroll (x, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndUnroll (x, v), exp)
                end
          | BndProj (x, i, v) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndProj (x, i, v), exp)
                end
          | BndNewTag (x, c) =>
                let val (x, exp) = alpha_vary (x, exp) replace_exp_exp
                in
                    (BndNewTag (x, c), exp)
                end


    (*** substitution for convars ***)

    fun subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                incon

    (*** derived forms ***)

    fun kndProd ks = foldr KndPair KndUnit ks

    fun conProd cs = foldr ConPair ConUnit cs

    fun conProj (n, c) =
        if n = 1
        then ConProj1 c
        else conProj (n-1, ConProj2 c)

    val tyUnit = TyProd []
    val tyBool = TySum [tyUnit, tyUnit]

    val tyInt = TyLit Literal.LitTyInt
    val tyChar = TyLit Literal.LitTyChar
    val tyString = TyLit Literal.LitTyString

    val valUnit = ValProd []
    fun valInt n = ValLit (Literal.LitInt n)
    fun valChar c = ValLit (Literal.LitChar c)
    fun valString s = ValLit (Literal.LitString s)

    fun valLam (x, con, exp) =
        let val f = Variable.newNamed "_"
        in
            ValFix (1, [(f, x, con, exp)])
        end
  end
